/*	License
 * 		inside EU: EUPL
 * 		outside EU: GPL or EUPL
 *  The program comes AS IS and without any warranty unless required by law
 *  (C) 2012 thecodemonkey
 */

package scheduler;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;



public class Scheduler2 implements IScheduler {
	protected List<SchedulerRunnable2> queued;
	protected List<SchedulerRunnable2> stopped;
	protected Map<Integer, SchedulerRunnable2> active;
	
	protected boolean running;
	protected int maxThreads, threadCounter;//, taskCounter, numberOfTasks;
	protected SchedulerClient client;
	
	public Scheduler2(SchedulerClient client, int maxThreads) {
		this.client=client;
		queued=Collections.synchronizedList(new LinkedList<SchedulerRunnable2>());
		stopped=Collections.synchronizedList(new LinkedList<SchedulerRunnable2>());
		active=Collections.synchronizedMap(new HashMap<Integer,SchedulerRunnable2>());
		this.maxThreads=maxThreads;
	}
	
	public synchronized void setMaxThreads(int maxThreads) {
		this.maxThreads = maxThreads;
	}
	
	@Override
	public void start() {
		synchronized (this) {
			running=true;
		}
		
		System.out.println("start");
		synchronized (queued) {
			int k=Math.min(queued.size(), maxThreads);
			
			for (int i=0;i<k;i++) {
				startTask();
			}
		}
		
	}

	@Override
	public int addTaskStopped(SchedulerRunnable2 t) {
		synchronized(stopped){
			stopped.add(t);
		}
		return t.getId();
	}
	 
	@Override
	public int addTask(SchedulerRunnable2 t) {
		synchronized(queued) {
			queued.add(t);
			
		}
		synchronized (this) {
			if (running) {
				if (threadCounter<maxThreads) {
					startTask();
				}
			}
		}
		return t.getId();
	}
	
	@Override
	public void stopAll() {
		
			synchronized (active) {
				running=false;
				Iterator<SchedulerRunnable2>it =active.values().iterator();
				while (it.hasNext()) {
					SchedulerRunnable2 t=it.next();
					t.stop();
				}
				synchronized (stopped) {
					stopped.addAll(active.values());
				}
				active.clear();
			}
			synchronized (queued) {
				running=false;
				Iterator<SchedulerRunnable2>it =queued.iterator();
				while (it.hasNext()) {
					SchedulerRunnable2 t=it.next();
					t.stop();
				}
				synchronized (stopped) {
					stopped.addAll(queued);
				}
				queued.clear();
			}
			
	}
	
	@Override
	public SchedulerRunnable2 remove(int id) {
		synchronized (active) {
			SchedulerRunnable2 t=active.get(id);
			
			if (t!=null) {
				t.stop();
				SchedulerRunnable2 thread = active.remove(id);
				synchronized (queued) {
					if (!queued.isEmpty()) {
						startTask();
					} else {
						synchronized (stopped) {
							if (active.isEmpty() &&stopped.isEmpty()) {
								running = false;
								client.allDone();
							}
						}
						
					}
						
				}
				
				return thread;
			} else {
				synchronized (queued) {
					for (SchedulerRunnable2 thread:queued) {
						if (thread.id==id) {
							queued.remove(thread);
							return thread;
						}
					}
				}
				synchronized (stopped) {
					for (SchedulerRunnable2 thread:stopped) {
						if (thread.id==id) {
							stopped.remove(thread);
							return thread;
						}
					}
				}
				
				return null;
			}
		}
		
	}
	
	@Override
	public void removeAll() {
		stopAll();
		queued.clear();
		stopped.clear();
	}
	
	@Override
	public SchedulerRunnable2 stop(int id) {
		synchronized (active){
			SchedulerRunnable2 t=active.remove(id);
			
			if (t!=null) {
				t.stop();
				stopped.add(t);
				if (active.isEmpty() && queued.isEmpty()) running=false;
			} else {
				synchronized (queued) {
					for (SchedulerRunnable2 thread:queued) {
						if (thread.id==id) {
							queued.remove(thread);
							synchronized (stopped) {
								stopped.add(thread);
							}
							return thread;
						}
					}
				}
			}
			return t;
		}
	}
	
	protected void startTask() {
		synchronized (this) {
			if (running) {
				SchedulerRunnable2 thread=queued.remove(0);
				synchronized (active) {
					active.put(thread.id,  thread);
				}
				
				threadCounter++; 
				new SchedulerThread(thread).start();
			}
		}
		
	}
	
	protected class SchedulerThread extends Thread {
		
		private SchedulerRunnable2 concreteTask;
		
		public SchedulerThread(SchedulerRunnable2 concreteTask) {
			this.concreteTask=concreteTask;
		}
		
		public SchedulerRunnable2 getConcreteTask() {
			return concreteTask;
		}
		
		@Override
		public void run() {
			concreteTask.run();
			synchronized (this) {
				threadCounter--;
			}
			
			synchronized (active) {
				active.remove(concreteTask.getId());
			}
			synchronized (queued) {
				if (!queued.isEmpty()) {
					startTask();
				} else {
					synchronized (active) {
						synchronized (stopped) {
							if (active.isEmpty() &&stopped.isEmpty()) {
								running = false;
								client.allDone();
							} 
						}
					}
					
				}
			}
			
		}
	}
	
	@Override
	public boolean isRunning() {
		synchronized (active) {
			return running;
		}
	}
	

	@Override
	public Collection<SchedulerRunnable2> getActive() {
		synchronized (active) {
			return new ArrayList<SchedulerRunnable2>(active.values());
			
		}
		
	}
	
	@Override
	public Collection<SchedulerRunnable2> getUnfinished() {
		synchronized (active) {
			List<SchedulerRunnable2> list=new ArrayList<SchedulerRunnable2>(active.size()+queued.size()+stopped.size());
			list.addAll(active.values());
			synchronized (queued) {
				list.addAll(queued);
			}
			
			synchronized (stopped) {
				list.addAll(stopped);
			}
			
			return list;
		}
	
	}

	/**
	 * Is active or queued
	 */
	@Override
	public boolean isRunning(int id) {
		synchronized (active) {
			if (active.containsKey(id)) {
				return true;
			}
		}
		synchronized (queued) {
			for (SchedulerRunnable2 sr:queued) {
				if (sr.getId()==id) return true;
			}
		}
		return false;
	}
	
	@Override
	public SchedulerRunnable2 enqueueStopped(int id) {
		synchronized (stopped) {
			Iterator<SchedulerRunnable2> it=stopped.iterator();
			SchedulerRunnable2 sr = null;
			while (it.hasNext()) {
				SchedulerRunnable2 task=it.next();
				if (task.getId()==id) {
					task.enqueue();
					synchronized (queued) {
						queued.add(0,task);
					}
					it.remove();
					sr=task;
					break;
				}
			}
			return sr;
		}
	}
	
	@Override
	public void enqueueAllStopped() {
		synchronized (stopped) {
			synchronized (queued) {
				for (SchedulerRunnable2 thread:stopped)
					thread.enqueue();
				queued.addAll(stopped);
				stopped.clear();
			}
		}
	}
	
	@Override
	public SchedulerRunnable2 getStoppedById(int id) {
		synchronized (stopped) {
			for (SchedulerRunnable2 r:stopped) {
				if (r.id == id) return r;
			}
			return null;
		}
	}
	
	@Override
	public Collection<SchedulerRunnable2> getStopped() {
		synchronized (stopped) {
			return new ArrayList<SchedulerRunnable2>(stopped);
		}
	}
	
//	@Override
//	public void reEnqueue(int id) {
//		synchronized (active) {
//			synchronized (queued) {
//				if (active.containsKey(id)) {
//					queued.add(active.remove(id));
//				}
//			}
//		}
//		
//	}
}
